# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for rotation_utils."""

from __future__ import annotations

from etils import enp
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
set_tnp = enp.testing.set_tnp


@enp.testing.parametrize_xnp()
def test_is_rotation_matrix(xnp: enp.NpModule):
  assert v3d.math.is_rot(xnp.eye(3))
  assert not v3d.math.is_rot(xnp.zeros((3, 3)))

  not_rot = xnp.array(
      [
          [0.0, 0.0, 1.0],
          [0.0, 1.0, 0.0],
          [0.0, 0.0, 1.0],
      ]
  )
  assert not v3d.math.is_rot(not_rot)

  delta = 1e-8
  rot = xnp.array(
      [
          [1.0 + delta, 0.0, 0.0],
          [0.0, 1.0, 0.0],
          [0.0, 0.0, 1.0],
      ]
  )
  assert v3d.math.is_rot(rot, atol=delta * 10)


@enp.testing.parametrize_xnp()
def test_is_rotation_matrix_raises(xnp: enp.NpModule):
  with pytest.raises(ValueError, match='Expected 3x3'):
    v3d.math.is_rot(xnp.eye(2))


@enp.testing.parametrize_xnp(with_none=True)
@pytest.mark.parametrize(
    'rad',
    [
        0.0,
        1 / 4 * enp.tau,
        -1 / 4 * enp.tau,
        1 / 2 * enp.tau,
        -1 / 2 * enp.tau,
        enp.tau,
        # Same shifted by 1 tau
        5 / 4 * enp.tau,
        -5 / 4 * enp.tau,
        3 / 2 * enp.tau,
        -3 / 2 * enp.tau,
        2 * enp.tau,
    ],
)
def test_rotation_around_axis(xnp: enp.NpModule, rad: float):
  if xnp is not None:
    rad = xnp.asarray(rad)
  else:
    xnp = enp.lazy.np
  rx = v3d.math.rot_x(rad)
  ry = v3d.math.rot_y(rad)
  rz = v3d.math.rot_z(rad)
  assert v3d.math.is_rot(rx)
  assert v3d.math.is_rot(ry)
  assert v3d.math.is_rot(rz)
  assert enp.lazy.get_xnp(rx) is xnp
  assert enp.lazy.get_xnp(ry) is xnp
  assert enp.lazy.get_xnp(rz) is xnp

  identity = xnp.eye(3)
  np.testing.assert_allclose(rx @ v3d.math.rot_x(-rad), identity)
  np.testing.assert_allclose(ry @ v3d.math.rot_y(-rad), identity)
  np.testing.assert_allclose(rz @ v3d.math.rot_z(-rad), identity)

  # Round trip euler <> matrix
  _assert_euler_round_trip(rx, xnp=xnp)
  _assert_euler_round_trip(ry, xnp=xnp)
  _assert_euler_round_trip(rz, xnp=xnp)


@enp.testing.parametrize_xnp(with_none=True)
@pytest.mark.parametrize(
    'rx, ry, rz',
    [
        (0.0, 0.0, 0.0),
        (1 / 4, -1 / 3, 1 / 5),
        (5 / 4, -4 / 3, 6 / 5),
    ],
)
def test_euler_roundtrip(xnp: enp.NpModule, rx: float, ry: float, rz: float):
  if xnp is not None:
    rx = xnp.asarray(rx)
    ry = xnp.asarray(ry)
    rz = xnp.asarray(rz)
  else:
    xnp = enp.lazy.np

  rot = v3d.math.euler_to_rot(x=rx, y=ry, z=rz)
  ax, ay, az = v3d.math.rot_to_euler(rot)
  rot_v2 = v3d.math.euler_to_rot(x=ax, y=ay, z=az)

  assert enp.lazy.get_xnp(rot) is xnp
  assert enp.lazy.get_xnp(rot_v2) is xnp
  np.testing.assert_allclose(rot, rot_v2, atol=1e-6, rtol=1e-7)


def _assert_euler_round_trip(rot, xnp):
  ax, ay, az = v3d.math.rot_to_euler(rot)
  assert enp.lazy.get_xnp(ax) is xnp
  assert enp.lazy.get_xnp(ay) is xnp
  assert enp.lazy.get_xnp(az) is xnp
  rot_v2 = v3d.math.euler_to_rot(x=ax, y=ay, z=az)
  assert enp.lazy.get_xnp(rot_v2) is xnp
  np.testing.assert_allclose(
      rot,
      rot_v2,
      atol=1e-6,
      rtol=1e-7,
      err_msg=f'For angles {(ax, ay, az)}',
  )



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation utils."""

from __future__ import annotations

import dataclasses
import functools
from typing import Callable, Generic, TypeVar

import dataclass_array as dca
from dataclass_array.typing import DcT
import einops
from etils import enp
from etils.array_types import FloatArray  # pylint: disable=g-multiple-import
from visu3d import array_dataclass
from visu3d import math
from visu3d.dc_arrays import ray as ray_lib
from visu3d.utils import np_utils
from visu3d.utils import py_utils
from visu3d.utils.lazy_imports import plotly_base

_T = TypeVar('_T')


class TransformBase(array_dataclass.DataclassArray):
  """Transformation base class.

  Defines the transformation interface.

  """

  def __matmul__(self, other: _T) -> _T:
    """Apply the transformation."""
    raise NotImplementedError('Abstract method.')

  @property
  def inv(self: _T) -> _T:
    """Returns the inverse camera transform."""
    raise NotImplementedError('Abstract method.')


@dataclasses.dataclass(frozen=True)
class Transform(TransformBase):
  """Affine transformation (Position, rotation and scale of an object).

  Attributes:
    R: Rotation/scale/skewing of the transformation ( `[[x0, y0, z0], [x1, y1,
      z1], [x2, y2, z2]]`)
    t: Translation of the transformation (`tx, ty, tz`)
  """

  R: FloatArray['*shape 3 3'] = (  # pylint: disable=invalid-name
      (1, 0, 0),
      (0, 1, 0),
      (0, 0, 1),
  )
  t: FloatArray['*shape 3'] = (0, 0, 0)

  @classmethod
  def identity(cls) -> Transform:
    """Returns the identity transform."""
    return cls(R=enp.lazy.np.eye(3), t=[0, 0, 0])

  @classmethod
  def from_matrix(cls, matrix: FloatArray['*shape 3 3']) -> Transform:
    """Constructs from a 4x4 transform matrix."""
    return cls(R=matrix[..., :3, :3], t=matrix[..., :3, 3])

  @classmethod
  def from_angle(cls, *, x=None, y=None, z=None) -> Transform:
    """Returns a transformation rotation around an axis (in radians).

    Example:

    ```python
    tr = v3d.Transform.from_angle(x=1/4 * enp.tau)  # Rotate 90° around x
    ```

    Rotations are applied following the Tait-Bryan chained rotations (z, y, x):

    ```python
    R = Rz @ Ry @ Rx
    ```

    See:
    https://en.wikipedia.org/wiki/Rotation_formalisms_in_three_dimensions#Euler_angles_(z-y%E2%80%B2-x%E2%80%B3_intrinsic)_%E2%86%92_rotation_matrix

    Args:
      x: Rotation around x (in radians): roll == ϕ == phi == x
      y: Rotation around y (in radians): pitch == θ == theta == y
      z: Rotation around z (in radians): yaw == ψ == psi == z

    Returns:
      tr: The transformation.
    """
    return cls(R=math.euler_to_rot(x=x, y=y, z=z))

  @classmethod
  @enp.check_and_normalize_arrays(strict=False)
  def from_look_at(
      cls,
      *,
      pos: FloatArray['*shape 3'],
      target: FloatArray['*shape 3'],
  ) -> Transform:
    """Factory to create a transformation which look at.

    Used to convert camera to world coordinates.

    This transformation assume the following conventions:

    * World coordinates: Floor is (x, y), z pointing upward
    * Camera coordinates: See `v3d.CameraSpec` docstring.

    The transformation assume the `width` dimension of the camera is parallel
    to the floor of the world.

    Args:
      pos: Origin position
      target: Target position

    Returns:
      The Transform.
    """
    return cls(
        t=pos,
        R=_get_r_look_at_(pos=pos, target=target),
    )

  @dca.vectorize_method
  def look_at(self, target: FloatArray['*shape 3']) -> Transform:
    """Returns a new transform looking at the target."""
    target = dca.utils.np_utils.asarray(
        target,
        xnp=self.xnp,
        dtype=enp.lazy.np.float32,
    )
    # TODO(epot): Rather than overwriting R, should only apply the rotation
    # to the existing R.
    return self.replace(R=_get_r_look_at_(pos=self.t, target=target))

  @property
  @dca.vectorize_method
  def x_dir(self) -> FloatArray['*shape 3']:
    """`x` axis of the transformation (`[x0, x1, x2]`)."""
    return self.R[:, 0]

  @property
  @dca.vectorize_method
  def y_dir(self) -> FloatArray['*shape 3']:
    """`y` axis of the transformation (`[y0, y1, y2]`)."""
    return self.R[:, 1]

  @property
  @dca.vectorize_method
  def z_dir(self) -> FloatArray['*shape 3']:
    """`z` axis of the transformation (`[z0, z1, z2]`)."""
    return self.R[:, 2]

  @property
  @dca.vectorize_method
  def x_ray(self) -> ray_lib.Ray:
    """Array pointing to `z`."""
    return ray_lib.Ray(pos=self.t, dir=self.x_dir)

  @property
  @dca.vectorize_method
  def y_ray(self) -> ray_lib.Ray:
    """Array pointing to `z`."""
    return ray_lib.Ray(pos=self.t, dir=self.y_dir)

  @property
  @dca.vectorize_method
  def z_ray(self) -> ray_lib.Ray:
    """Array pointing to `z`."""
    return ray_lib.Ray(pos=self.t, dir=self.z_dir)

  @property
  @dca.vectorize_method
  def ray_basis(self) -> ray_lib.Ray:
    """The `(x, y, z)` basis of the transformation, as ray."""
    return ray_lib.Ray(
        # We can use `np` for the display
        pos=self.xnp.broadcast_to(self.t, self.shape + (3, 3)),
        # R is [[x0, y0, z0], [x1, y1, z1], [x2, y2, z2]] so we transpose
        # so dir is [[x0, x1, x2], [y0, y1, y2], [z0, z1, z2]]
        dir=self.xnp.asarray(self.R.T),
    )

  @property
  @dca.vectorize_method
  def scale_xyz(self) -> FloatArray['*shape 3']:
    """Returns the `(sx, sy, sz)` scale of the transform along each axis."""
    return enp.linalg.norm(self.R, axis=0)

  @property
  @dca.vectorize_method
  def scale(self):
    """Returns the global scale (if `x, y, z` share the same scale).

    Returns:
      The global scale shared by all axis.

    Raises:
      ValueError: If the `x, y, z` scales are different between axis.
    """

    def _err_msg():
      return (
          'Cannot get `Transform.scale` when x, y, z scale are '
          f'different: {self.scale_xyz}'
      )

    xnp = self.xnp
    scale_xyz = xnp.round(self.scale_xyz, decimals=7)
    # TODO(epot): Move into `enp.ops.unique`
    if xnp is enp.lazy.np:
      global_scales = xnp.unique(scale_xyz)
      raise_error = len(global_scales) != 1
    elif xnp is enp.lazy.jnp:
      global_scales, global_count = xnp.unique(
          scale_xyz,
          size=1,
          return_counts=True,
      )

      # Unfortunately, Jax don't have an easy API to conditionally
      # raise error within a `jax.jit` function
      # This won't have any effect when the function is traced.
      from jax.experimental import checkify  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top

      checkify.check(global_count[0] == 3, msg=_err_msg())

      raise_error = False
    elif xnp is enp.lazy.tnp:
      global_scales, _ = enp.lazy.tf.unique(scale_xyz)
      raise_error = len(global_scales) != 1
    else:
      raise AssertionError(f'Unknown numpy module {xnp}')
    if raise_error:
      raise ValueError(_err_msg())
    else:
      (global_scale,) = global_scales
    return global_scale

  def mul_scale(self, factor: FloatArray['*shape 3?']) -> Transform:
    """Scale the transformation `(sx, sy, sz)` by the given factor.

    This is similar to `v3d.Ray.scale_dir`, applied on each individual axis.

    Args:
      factor: The factor by which scale each axis. If scalar, broadcast all axis
        by the same value.

    Returns:
      The new transform with scaled.
    """
    xnp = self.xnp
    factor = xnp.asarray(factor)
    if factor.shape == ():  # pylint: disable=g-explicit-bool-comparison
      new_r = self.R * factor
    elif factor.shape == (3,):
      new_r = self.R @ xnp.diag(factor)
    else:
      raise ValueError(f'Unsupported factor shape {factor.shape}')
    return self.replace(R=new_r)

  @dca.vectorize_method
  def normalize(self) -> Transform:
    """Normalize the scale x, y, z to be `(1, 1, 1)`."""
    xnp = self.xnp
    return self.replace(R=self.R @ xnp.diag(1 / self.scale_xyz))

  @property
  @dca.vectorize_method
  def matrix4x4(self) -> FloatArray['*shape 4 4']:
    """Returns the 4x4 transformation matrix.

    [R|t]
    [0|1]
    """
    t = einops.rearrange(self.t, '... d -> ... d 1')
    matrix3x4 = self.xnp.concatenate([self.R, t], axis=-1)
    assert matrix3x4.shape == (3, 4)
    last_row = self.xnp.asarray([[0, 0, 0, 1]])
    return self.xnp.concatenate([matrix3x4, last_row], axis=-2)

  @property
  @dca.vectorize_method
  def inv(self) -> Transform:
    """Returns the inverse camera transform."""
    # Might be a more optimized way than stacking/unstacking matrix
    return type(self).from_matrix(enp.linalg.inv(self.matrix4x4))

  def __add__(self, translation: FloatArray['... 3']) -> Transform:
    """Translate the position."""
    translation = self.xnp.asarray(translation)
    return self.replace(t=self.t + translation)

  __sub__ = np_utils.__sub__

  @dca.vectorize_method
  def __matmul__(self, other: _T) -> _T:
    """Apply the transformation the array or dataclass array.

    Transformation & inputs can be arbitrarly batched. Transform will be
    applied on individual elements (vectorized, NOT broadcasted):

    For example, for `tr @ other`, the shape are vectorized using the
    `dca.vectorize_method` rules:

    ```python
    () @ (*x,) -> (*x,)
    (*tr,) @ (*tr, *x) -> (*tr, *x)
    ```

    Any `v3d.DataclassArray` can support `v3d.Transform` by implementing
    the `apply_transform` protocol:

    ```python
    my_obj = tr @ my_obj  # Call `my_obj.apply_transform(tr)`
    ```

    Inside the `apply_transform` function, `tr.shape == ()`. Vectorization is
    auto-supported.

    Args:
      other: The array or dataclass array on which apply the transformation.
        Arrays are interpreted as 3d point cloud so should be `Array[..., 3]`.
        `v3d.DataclassArray` should implement the `apply_transform` protocol.

    Returns:
      The new `other` object after transformation.
    """
    self.assert_same_xnp(other)
    if enp.lazy.is_array(other):
      return self.apply_to_pos(other)
    elif isinstance(other, dca.DataclassArray):
      py_utils.assert_supports_protocol(
          other,
          'apply_transform',
          msg=f'`v3d.Transform` @ `{type(other).__name__}` not supported.',
      )
      return other.apply_transform(self)
    else:
      raise TypeError(f'Unexpected type: {type(other)}')

  # TODO(epot): Also add a `tr.apply_to` method which supports broadcasting

  @dca.vectorize_method
  def apply_to_pos(self, point: FloatArray['*d 3']) -> FloatArray['*d 3']:
    """Apply the transformation on the point cloud."""
    self.assert_same_xnp(point)
    point = self.xnp.asarray(point)
    if point.shape[-1] != 3:
      raise ValueError(f'point shape should be `(..., 3)`. Got {point.shape}')
    return self.apply_to_dir(point) + self.t

  @dca.vectorize_method
  def apply_to_dir(
      self,
      direction: FloatArray['*d 3'],
  ) -> FloatArray['*d 3']:
    """Apply the transformation on the direction."""
    self.assert_same_xnp(direction)
    # Direction are invariant to translation
    return self.xnp.einsum('ij,...j->...i', self.R, direction)

  # Protocols (inherited)

  # Apply transform is the protocol implementation but is NOT part of the public
  # API, so don't have `@dca.vectorize_method`
  def apply_transform(
      self,
      tr: Transform,
  ) -> Transform:
    return self.replace(
        R=tr.R @ self.R,
        t=tr.apply_to_pos(self.t),
    )

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    rays = self.ray_basis
    (line_trace,) = rays.make_traces()
    line_trace.mode = 'lines+markers+text'
    # Add x, y, z text labels to the plot
    # Each point in the original line trace is:
    # [ray_origin, ray_end, None(=line break)]
    # So we only add the text on the `ray_end`
    # fmt: off
    line_trace.text = [
        None, 'x', None,
        None, 'y', None,
        None, 'z', None,
    ] * rays.size
    # fmt: on
    return [line_trace]


# TODO(epot): Have custom transform support `.inv`
#
# @cam_from_px.register_inv
# def px_from_cam(self, point3d):
#


def custom_transform(
    fn: Callable[[DcT, _T], _T],
) -> Callable[[DcT], CustomTransform[DcT, _T]]:
  """Custom transformation decorator.

  This decorator is a wrapper around custom transformations method to allow
  composing custom transformations with `v3d.Transform`.

  Usage:

  ```python
  class MyObject(v3d.DataclassArray):

    @property
    @v3d.custom_transform
    def my_transform(self, points: _T) -> _T:
      ...  # Custom transform implementation
      return points

  obj = MyObject()

  # The custom transform can be composed with other
  points = obj.my_transform @ points
  points = obj.my_transform @ cam.world_from_px @ px_coords
  ```

  Note: The decorated method is automatically vectorized. See
  `dca.vectorize_method`.

  Args:
    fn: Method to decorate

  Returns:
    The decorated function (which is should be decorated with @property).
  """

  @functools.wraps(fn)
  def decorated(self) -> CustomTransform:
    return CustomTransform(self_=self, method=fn)

  return decorated


@dataclasses.dataclass(frozen=True)
class CustomTransform(TransformBase, Generic[DcT, _T]):  # pytype: disable=invalid-function-definition
  """Custom transformation method wrapper.

  Attributes:
    self_: Object from which the transform is applied.
    method: Transform method (decorated by `@v3d.custom_transform`). Calling
      `custom_tr @ x` is equivalent of calling `self.method(x)`
  """

  self_: DcT = dca.field(  # pytype: disable=annotation-type-mismatch
      shape=(),
      dtype=dca.DataclassArray,
  )
  # It is important that:
  # * method should NOT be a lambda (to avoid accidental closure side effects
  # * method should NOT be a bound method (to not bind `self`)
  # TODO(epot): Add `__post_init__` check for this.
  method: Callable[[DcT, _T], _T]

  @dca.vectorize_method
  def __matmul__(self, other: _T) -> _T:
    if isinstance(other, TransformBase):
      # If other is another transform: create a composed transform
      return self.compose_transform(other)
    else:
      return self.method(self.self_, other)

  def compose_transform(self, other_tr: TransformBase) -> TransformBase:
    """Called when `tr_custom @ tr`."""
    return CustomTransform(
        self_=ComposedTransform(
            left_tr=self,
            right_tr=other_tr,
        ),
        method=ComposedTransform._apply_to,  # pylint: disable=protected-access
    )

  def apply_transform(self, tr: Transform) -> DcT:
    """Called when `tr @ tr_custom`."""
    return CustomTransform(
        self_=ComposedTransform(
            left_tr=tr,
            right_tr=self,
        ),
        method=ComposedTransform._apply_to,  # pylint: disable=protected-access
    )


@dataclasses.dataclass(frozen=True)
class ComposedTransform(TransformBase):
  """Transform composed to 2 transformations.

  Allow to chain arbitrary custom transforms:

  ```
  px = px_from_cam @ cam_from_world @ world_points
  ```

  Note: Should not be created directly but only inside `CustomTransform`.

  """

  left_tr: TransformBase
  right_tr: TransformBase

  def _apply_to(self: ComposedTransform, other: _T) -> _T:
    """Apply composed method."""
    return self.left_tr @ (self.right_tr @ other)


def _get_r_look_at_(
    *,
    pos: FloatArray['*shape 3'],
    target: FloatArray['*shape 3'],
) -> FloatArray['*shape 3 3']:
  """Compute the `R` (3, 3) matrix."""
  # TODO(epot): Support more conventions
  # * `up='z'`
  # * `mode='camera'` (x==h pointing ↓, y==w pointing →) or right hand rule.

  xnp = enp.lazy.get_xnp(pos)
  _assert_shape(pos, 'pos')
  _assert_shape(target, 'target')

  cam_forward = enp.linalg.normalize(target - pos)

  # In world coordinates, `z` is pointing up
  world_up = xnp.array([0, 0, 1])
  # The width of the cam is parallel to the ground (prependicular to z), so
  # use cross-product.
  cam_w = xnp.cross(cam_forward, world_up)
  cam_w = enp.linalg.normalize(cam_w)

  # Similarly, the height is pointing downward.
  cam_h = xnp.cross(cam_forward, cam_w)

  R = xnp.stack([cam_w, cam_h, cam_forward], axis=-1)  # pylint: disable=invalid-name
  return R


def _assert_shape(array: FloatArray['*d 4'], name: str) -> None:
  """Test that array shape end by 3."""
  if array.shape[-1] != 3:
    raise ValueError(f'{name!r} shape should end be (3,). Got {array.shape}')



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy imports."""

from __future__ import annotations

import typing

from visu3d.utils import py_utils

if typing.TYPE_CHECKING:
  # Plotly is an optional (but strongly recommended on colab) dependency
  import IPython  # pytype: disable=import-error
  from plotly import graph_objects as plotly_go
  from plotly import basedatatypes as plotly_base
  import scipy
else:
  # Hack: Use `globals` as a hack to avoid VS Code to infer go as
  # `go | LazyModule`
  globals().update(
      IPython=py_utils.LazyModule('IPython'),
      plotly_go=py_utils.LazyModule('plotly.graph_objects'),
      plotly_base=py_utils.LazyModule('plotly.basedatatypes'),
      scipy=py_utils.LazyModule('scipy'),
  )



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for plotly."""

from __future__ import annotations

import chex
import dataclass_array as dca
from etils import enp
import numpy as np
import pytest
import visu3d as v3d
from visu3d.utils.lazy_imports import plotly_base
from visu3d.utils.lazy_imports import plotly_go as go


@enp.testing.parametrize_xnp()
def test_to_xyz_dict(xnp: enp.NpModule):
  chex.assert_tree_all_close(
      v3d.plotly.to_xyz_dict(
          [
              [0, 1, 2],
              [0, 10, 20],
              [0, 100, 200],
          ]
      ),
      {
          'x': xnp.array([0, 0, 0]),
          'y': xnp.array([1, 10, 100]),
          'z': xnp.array([2, 20, 200]),
      },
  )

  chex.assert_tree_all_close(
      v3d.plotly.to_xyz_dict(
          [
              [0, 1, 2],
              [0, 10, 20],
              [0, 100, 200],
          ],
          pattern='axis_{}',
          names='uvw',
          axis=0,  # pytype: disable=wrong-arg-types
      ),
      {
          'axis_u': xnp.array([0, 1, 2]),
          'axis_v': xnp.array([0, 10, 20]),
          'axis_w': xnp.array([0, 100, 200]),
      },
  )


class VisuObj(v3d.plotly.Visualizable):
  """Test object."""

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    """Construct the traces of the given object."""
    return [
        go.Scatter3d(
            x=[0, 1, 2],
            y=[0, 1, 2],
            z=[0, 1, 2],
        ),
    ]


class VisuObjImplicit:
  """Test object which do not inherit from `v3d.plotly.Visualizable`."""

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    """Construct the traces of the given object."""
    return [
        go.Scatter3d(
            x=[0, 1, 2],
            y=[0, 1, 2],
            z=[0, 1, 2],
        ),
    ]


class VisuObjInvalid:
  """Not a visualizable."""


def test_is_visualizable():
  assert v3d.plotly.fig_utils.is_visualizable(VisuObj())
  assert v3d.plotly.fig_utils.is_visualizable(VisuObjImplicit())
  assert not v3d.plotly.fig_utils.is_visualizable(VisuObjInvalid())


def test_make_fig():
  x_trace = go.Scatter3d(
      x=[0, 1, 2],
      y=[0, 1, 2],
      z=[0, 1, 2],
  )
  x_array = np.ones((4, 3))
  fig = v3d.make_fig(
      [
          VisuObj(),
          VisuObjImplicit(),
          x_trace,
          x_array,
      ]
  )
  assert isinstance(fig, go.Figure)

  with pytest.raises(TypeError, match='Unsuported'):
    v3d.make_fig([VisuObjInvalid()])


@pytest.mark.parametrize('shape', [(), (4, 3), (700,), (3, 100, 4)])
def test_subsample(shape: dca.typing.Shape):
  r = v3d.Ray(pos=np.ones(shape + (3,)), dir=np.ones(shape + (3,)))
  _ = r.fig

  r = v3d.Point3d(p=np.ones(shape + (3,)), rgb=np.ones(shape + (3,)))
  _ = r.fig



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for np_utils."""

from __future__ import annotations

from etils import enp
import numpy as np
import visu3d as v3d
from visu3d.utils import np_utils


@enp.testing.parametrize_xnp()
def test_append_row(xnp: enp.NpModule):
  x = xnp.ones((2, 4))
  y = np_utils.append_row(x, value=4.0, axis=-1)
  assert isinstance(y, xnp.ndarray)
  expected = [
      [1, 1, 1, 1, 4],
      [1, 1, 1, 1, 4],
  ]
  np.testing.assert_allclose(y, expected)

  y = np_utils.append_row(x, value=4.0, axis=0)
  assert isinstance(y, xnp.ndarray)
  expected = [
      [1, 1, 1, 1],
      [1, 1, 1, 1],
      [4, 4, 4, 4],
  ]
  np.testing.assert_allclose(y, expected)


def test_interpolate():
  p = [
      [0, 0, 0],
      [0, 2, 0],
      [0, 2, 1],
      [0, 0, -5],
  ]
  p_interp = v3d.math.interp_points(p, t=100)
  assert p_interp.shape == (100, 3)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for camera_spec."""

from __future__ import annotations

import dataclass_array as dca
from etils import enp
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
set_tnp = enp.testing.set_tnp

H, W = 640, 480


def make_camera_spec(
    *,
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
) -> v3d.PinholeCamera:
  spec = v3d.PinholeCamera.from_focal(
      resolution=(H, W),
      focal_in_px=35.0,
      xnp=xnp,
  )
  spec = spec.broadcast_to(shape)
  return spec


@enp.testing.parametrize_xnp(with_none=True)
@pytest.mark.parametrize('spec_shape', [(), (3,), (2, 3)])
def test_camera_spec_init(
    xnp: enp.NpModule,
    spec_shape: dca.typing.Shape,
):
  if spec_shape and xnp is enp.lazy.tnp:
    pytest.skip('Vectorization not supported yet with TF')

  spec = make_camera_spec(xnp=xnp, shape=spec_shape)
  assert spec.resolution == (H, W)
  assert spec.h == H
  assert spec.w == W
  assert spec.wh == (W, H)
  assert spec.hw == (H, W)
  assert spec.shape == spec_shape
  assert spec.K.shape == spec_shape + (3, 3)

  if xnp is None:
    xnp = np
  assert spec.xnp is xnp

  x = _broadcast_to(xnp, [0, 0, 1], (1,) * len(spec_shape) + (3,))
  assert isinstance(spec.px_from_cam @ x, xnp.ndarray)

  x = _broadcast_to(xnp, [0, 0], (1,) * len(spec_shape) + (2,))
  assert isinstance(spec.cam_from_px @ x, xnp.ndarray)

  _ = spec.fig

  spec = spec.replace_fig_config(scale=3.0)
  _ = spec.fig


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('spec_shape', [(), (3,), (2, 3)])
@pytest.mark.parametrize('point_shape', [(), (2, 3)])
def test_camera_spec_central_point(
    xnp: enp.NpModule,
    spec_shape: dca.typing.Shape,
    point_shape: dca.typing.Shape,
):
  if spec_shape and xnp is enp.lazy.tnp:
    pytest.skip('Vectorization not supported yet with TF')

  spec = make_camera_spec(xnp=xnp, shape=spec_shape)

  # Projecting the central point (batched)
  central_point_cam = _broadcast_to(
      xnp,
      [0, 0, 1],
      spec_shape + point_shape + (3,),
  )
  central_point_px = spec.px_from_cam @ central_point_cam
  assert isinstance(central_point_px, xnp.ndarray)
  assert central_point_px.shape == spec_shape + point_shape + (2,)
  np.testing.assert_allclose(
      central_point_px,
      np.broadcast_to([W / 2, H / 2], spec_shape + point_shape + (2,)),
  )

  # Round trip conversion
  np.testing.assert_allclose(
      central_point_cam,
      spec.cam_from_px @ central_point_px,
      atol=1e-4,
  )


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('spec_shape', [(), (2, 3)])
def test_camera_px_centers(
    xnp: enp.NpModule,
    spec_shape: dca.typing.Shape,
):
  if spec_shape and xnp is enp.lazy.tnp:
    pytest.skip('Vectorization not supported yet with TF')

  spec = make_camera_spec(xnp=xnp, shape=spec_shape)

  px_centers = spec.px_centers()
  assert isinstance(px_centers, xnp.ndarray)
  assert px_centers.shape == spec_shape + (H, W, 2)

  cam_centers = spec.cam_from_px @ px_centers
  assert isinstance(cam_centers, xnp.ndarray)
  assert cam_centers.shape == spec_shape + (H, W, 3)

  round_trip_px = spec.px_from_cam @ cam_centers
  assert isinstance(round_trip_px, xnp.ndarray)
  assert round_trip_px.shape == spec_shape + (H, W, 2)

  np.testing.assert_allclose(round_trip_px, px_centers, atol=1e-4)
  # Scaling/normalizing don't change the projection
  np.testing.assert_allclose(
      round_trip_px,
      spec.px_from_cam @ (cam_centers * 3.0),
      atol=1e-4,
  )


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('spec_shape', [(), (2, 3)])
def test_camera_points(
    xnp: enp.NpModule,
    spec_shape: dca.typing.Shape,
):
  if spec_shape and xnp is enp.lazy.tnp:
    pytest.skip('Vectorization not supported yet with TF')

  spec = make_camera_spec(xnp=xnp, shape=spec_shape)

  # Random point cloud in camera coordinates
  rng = np.random.default_rng(0)
  coords3d = rng.random(spec_shape + (5, 3))
  rgb = rng.integers(255, size=spec_shape + (5, 3))
  points3d = v3d.Point3d(p=coords3d, rgb=rgb)
  points3d = points3d.as_xnp(xnp)

  px = spec.px_from_cam @ points3d
  assert isinstance(px, v3d.Point2d)
  assert px.shape == spec_shape + (5,)
  assert px.depth is not None

  # Round-trip should be a no-op (as depth is preserved)
  round_trip_points3d = spec.cam_from_px @ px
  assert isinstance(round_trip_points3d, v3d.Point3d)
  assert round_trip_points3d.shape == spec_shape + (5,)
  dca.testing.assert_allclose(round_trip_points3d, points3d, atol=1e-5)

  # Round-trip with depth=None project to z=1
  px = px.replace(depth=None)
  round_trip_points3d = spec.cam_from_px @ px
  assert isinstance(round_trip_points3d, v3d.Point3d)
  assert round_trip_points3d.shape == spec_shape + (5,)
  np.testing.assert_allclose(
      round_trip_points3d.p[..., -1],
      np.ones(spec_shape + (5,)),
      # atol=1e-4,
  )


def _broadcast_to(xnp: enp.NpModule, array, shape):
  return xnp.broadcast_to(xnp.array(array), shape)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Set plotly trace names."""

from __future__ import annotations

import collections
import dataclasses
import itertools

from etils import enp
from visu3d.plotly import fig_utils
from visu3d.utils.lazy_imports import plotly_base


@dataclasses.dataclass
class TraceNamer:
  """Set plotly trace names."""

  name_to_count: dict[str, itertools.count] = dataclasses.field(
      default_factory=lambda: collections.defaultdict(itertools.count),
  )

  def set_name(
      self,
      traces: list[plotly_base.BaseTraceType],
      array: fig_utils.VisualizableInterface,
  ) -> None:
    """Update the name of the traces."""
    if not traces:  # No traces
      return

    elif fig_utils.is_visualizable(array):
      if fig_utils.has_fig_config(array) and array.fig_config.name:
        # User-defined name
        name = array.fig_config.name
      else:
        name = type(array).__name__
    elif enp.lazy.is_array(array):
      name = 'points'
    else:
      raise TypeError(f'Unexpected trace {type(array)}')

    curr_id = next(self.name_to_count[name])

    # Only update the name of the first trace
    # In the future, could have more complicated heuristics based on cases
    for trace in traces:
      if not trace.name:
        trace.name = f'{name} {curr_id}'
      # Set the group so all items are triggered together
      trace.legendgroup = f'{name} {curr_id}'



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Rotation utils.

Attributes:
  DEG2RAD: Constant to convert degrees to radians.
  RAD2DEG: Constant to convert radians to degrees.
"""

from __future__ import annotations

from typing import Tuple

from etils import enp
from etils.array_types import FloatArray

DEG2RAD = enp.tau / 360.0
RAD2DEG = 360.0 / enp.tau

# TODO(epot): Support vectorization

# Could eventually add utils to combine rotations together, like:
# yaw == ψ == psi == z
# pitch == θ == theta == y
# roll == ϕ == phi == x
# But might be better to let users explicitly compose transforms.


@enp.check_and_normalize_arrays(strict=False)
def rot_x(angle: FloatArray[''], xnp: enp.NpModule = ...) -> FloatArray['3 3']:
  """Rotation matrix for rotation around X (in radians)."""
  if angle.ndim:
    raise ValueError(f'Rotation angle should be scalar. Not {angle.shape}')
  c = xnp.cos(angle)
  s = xnp.sin(angle)
  R = xnp.array(  # pylint: disable=invalid-name
      [
          [1, 0, 0],
          [0, c, -s],
          [0, s, c],
      ]
  )
  return R


@enp.check_and_normalize_arrays(strict=False)
def rot_y(angle: FloatArray[''], xnp: enp.NpModule = ...) -> FloatArray['3 3']:
  """Rotation matrix for rotation around Y (in radians)."""
  if angle.ndim:
    raise ValueError(f'Rotation angle should be scalar. Not {angle.shape}')
  c = xnp.cos(angle)
  s = xnp.sin(angle)
  R = xnp.array(  # pylint: disable=invalid-name
      [
          [c, 0, s],
          [0, 1, 0],
          [-s, 0, c],
      ]
  )
  return R


@enp.check_and_normalize_arrays(strict=False)
def rot_z(angle: FloatArray[''], xnp: enp.NpModule = ...) -> FloatArray['3 3']:
  """Rotation matrix for rotation around Z (in radians)."""
  if angle.ndim:
    raise ValueError(f'Rotation angle should be scalar. Not {angle.shape}')
  c = xnp.cos(angle)
  s = xnp.sin(angle)
  R = xnp.array(  # pylint: disable=invalid-name
      [
          [c, -s, 0],
          [s, c, 0],
          [0, 0, 1],
      ]
  )
  return R


def euler_to_rot(
    x=None,
    y=None,
    z=None,
    *,
    order: str = 'zyx',
) -> FloatArray['3 3']:
  """Creates a 3x3 matrix from the euler radian angles.

  By default, rotations are applied following the Tait-Bryan chained rotations
  (z, y, x):

  ```python
  R = Rz @ Ry @ Rx
  ```

  See:
  https://en.wikipedia.org/wiki/Rotation_formalisms_in_three_dimensions#Euler_angles_(z-y%E2%80%B2-x%E2%80%B3_intrinsic)_%E2%86%92_rotation_matrix

  Args:
    x: Rotation around x (in radians): roll == ϕ == phi == x
    y: Rotation around y (in radians): pitch == θ == theta == y
    z: Rotation around z (in radians): yaw == ψ == psi == z
    order: Axis order convention used (e.g. `'xyz'`, `'zyx'`,...)

  Returns:
    tr: The transformation.
  """
  order = tuple(order)

  if set(order) != set('xyz') or len(order) != 3:
    raise ValueError(
        f'Order should contain x, y, z exactly once. Got {order!r}'
    )

  def _accumulate_rot(r0, r1):
    return r1 if r0 is None else r0 @ r1

  axis_to_val = {'x': x, 'y': y, 'z': z}
  axis_to_rot_fn = {
      'x': rot_x,
      'y': rot_y,
      'z': rot_z,
  }

  r_final = None
  for axis in order:
    angle = axis_to_val[axis]
    if angle is None:
      continue
    r = axis_to_rot_fn[axis](angle)
    r_final = _accumulate_rot(r_final, r)

  if r_final is None:  # All x, y, z undefined => Identity
    r_final = enp.lazy.np.eye(3)
  return r_final


@enp.check_and_normalize_arrays
def rot_to_euler(
    rot: FloatArray['3 3'],
    *,
    eps: float = 1e-6,
    xnp: enp.NpModule = ...,
) -> Tuple[FloatArray[''], FloatArray[''], FloatArray['']]:
  """Extract euler angles from a 3x3 rotation matrix.

  Like `euler_to_rot`, it follow the z, y, x convension, BUT returns x, y, z.

  From: https://www.geometrictools.com/Documentation/EulerAngles.pdf

  Args:
    rot: Rotation matrix
    eps: Precision threshold to detect 90 degree angles.
    xnp: Np module used (jnp, tnp, np,...)

  Returns:
    The x, y, z euler angle (in radian)
  """
  r00 = rot[0, 0]
  # r01 = rot[0, 1]
  r02 = rot[0, 2]
  r10 = rot[1, 0]
  r11 = rot[1, 1]
  r12 = rot[1, 2]
  r20 = rot[2, 0]
  r21 = rot[2, 1]
  r22 = rot[2, 2]

  if xnp.abs(r20) < 1.0 - eps:  # Should allow to tune the precision ?
    theta_y = xnp.arcsin(-r20)
    theta_z = xnp.arctan2(r10, r00)
    theta_x = xnp.arctan2(r21, r22)
  else:  # r20 == +1 / -1
    sign = +1 if r02 > 0 else -1

    theta_y = sign * enp.tau / 4
    theta_z = -sign * xnp.arctan2(-r12, r11)
    theta_x = 0.0

  theta_x = xnp.asarray(theta_x)
  theta_y = xnp.asarray(theta_y)
  theta_z = xnp.asarray(theta_z)
  return theta_x, theta_y, theta_z


@enp.check_and_normalize_arrays
def is_orth(
    rot: FloatArray['3 3'],
    *,
    atol: float = 1e-6,
    xnp: enp.NpModule = ...,
) -> bool:
  """Check if the matrix is a valid orthogonal matrix `O(3)`.

  Each orthogonal matrix form a orthonormal basis.
  The group of all 3x3 orthogonal matrices consists of all proper and improper
  rotations (reflexions).

  See:
  https://en.wikipedia.org/wiki/3D_rotation_group#Orthogonal_and_rotation_matrices

  Args:
    rot: The 3x3 matrix to check
    atol: Precision at which checking the matrix
    xnp: Np module used (jnp, tnp, np,...)

  Returns:
    True if the matrix is orthogonal.
  """
  if rot.shape != (3, 3):
    raise ValueError(f'Expected 3x3 shape, but got {rot.shape}')

  should_be_identity = rot.T @ rot
  identity = xnp.eye(3, dtype=rot.dtype)
  diff = enp.linalg.norm(identity - should_be_identity)
  return diff < atol


def is_rot(rot: FloatArray['3 3'], *, atol: float = 1e-6) -> bool:
  """Checks if a matrix is a valid rotation matrix `SO(3)`.

  This is done by checking:

  * R.T @ R = Identity (R is orthogonal)
  * det(R) = +1 (-1 corresponding to reflexions)

  If you don't care about reflexions, use `is_orth` instead.

  See:
  https://en.wikipedia.org/wiki/3D_rotation_group#Orthogonal_and_rotation_matrices

  Args:
    rot: The 3x3 matrix to check
    atol: Precision at which checking the matrix

  Returns:
    True if the matrix is a rotation matrix.
  """
  if rot.shape != (3, 3):
    raise ValueError(f'Expected 3x3 shape, but got {rot.shape}')

  det = enp.linalg.det(rot)
  return is_orth(rot, atol=atol) and det > 0.0



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ray utils."""

from __future__ import annotations

import dataclasses

import dataclass_array as dca
from dataclass_array.typing import Axes
from etils import enp
from etils.array_types import Array, FloatArray  # pylint: disable=g-multiple-import
from visu3d import array_dataclass
from visu3d import plotly
from visu3d.dc_arrays import transformation
from visu3d.plotly import fig_config_utils
from visu3d.utils import np_utils
from visu3d.utils.lazy_imports import plotly_base

# TODO(epot):
# * Make the dir optional to allow:
#   ray = Ray(pos=[0, 0, 0]).look_at(target)
# * Check broadcasting for `+` and `*`


@dataclasses.dataclass(frozen=True)
class Ray(array_dataclass.DataclassArray):
  """6d vector with position and direction.

  Note: The direction is not normalized by default.

  Attributes:
    pos: Position
    dir: Direction
  """

  pos: FloatArray['*shape 3']
  dir: FloatArray['*shape 3']

  # Overwrite `v3d.DataclassArray.fig_config`.
  fig_config: plotly.TraceConfig = dataclasses.field(
      default=plotly.TraceConfig(
          num_samples=fig_config_utils.LazyValue(
              lambda: fig_config_utils.fig_config.num_samples_ray
          )
      ),
      repr=False,
      init=False,
  )

  @property
  def end(self) -> FloatArray['*shape 3']:
    """The extremity of the ray (`ray.pos + ray.dir`)."""
    return self.pos + self.dir

  @classmethod
  @enp.check_and_normalize_arrays(strict=False)
  def from_look_at(
      cls,
      *,
      pos: Array['*d 3'],
      target: Array['*d 3'],
  ) -> Ray:
    """Factory to create a look at Ray.

    Alias of `Ray(pos=pos, dir=target-from)`.

    Args:
      pos: Origin position
      target: Target position

    Returns:
      The Ray.
    """
    return cls(
        pos=pos,
        dir=target - pos,
    )

  def __add__(self, translation: FloatArray['... 3']) -> Ray:
    """Translate the position."""
    if isinstance(translation, Ray):
      raise TypeError(
          'Cannot add Ray with Ray. '
          'In `ray + x`: x should be a FloatArray[..., 3].'
      )
    translation = self.xnp.asarray(translation)
    return self.replace(pos=self.pos + translation)

  __sub__ = np_utils.__sub__

  def scale_dir(self, scale: FloatArray['...']) -> Ray:
    """Scale the dir."""
    if isinstance(scale, Ray):
      raise TypeError(
          'Cannot multiply Ray with Ray. '
          'In `ray * x`: x should be a scalar factor.'
      )
    scale = self.xnp.asarray(scale)
    return self.replace(dir=self.dir * scale)

  def norm(self, keepdims: bool = False) -> FloatArray['*shape']:
    """Returns the norm of the dir."""
    return enp.linalg.norm(self.dir, axis=-1, keepdims=keepdims)

  def normalize(self) -> Ray:
    """Normalize the directions."""
    return self.replace(dir=enp.linalg.normalize(self.dir))

  def mean(self, *, axis: Axes = None) -> Ray:
    """Returns the average ray."""
    # Mean reduce across all axis but the last one
    axis = self._to_absolute_axis(axis)
    return self.map_field(lambda t: t.mean(axis=axis))

  def look_at(self, target: Array['*shape 3']) -> Ray:
    """Change the direction to point to the target point."""
    # Could add a `keep_norm=True` ?
    target = dca.utils.np_utils.asarray(target, xnp=self.xnp)
    return self.replace(dir=target - self.pos)

  # Protocols (inherited)

  def apply_transform(self, tr: transformation.Transform) -> Ray:
    return self.replace(
        pos=tr.apply_to_pos(self.pos),
        dir=tr.apply_to_dir(self.dir),
    )

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    start = self.pos
    end = self.end
    return plotly.make_lines_traces(
        start=start,
        end=end,
        end_marker='diamond',
        axis=-1,
    )



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Camera spec utils."""

from __future__ import annotations

import abc
import dataclasses
import functools
import typing
from typing import Any, Optional, Tuple

import dataclass_array as dca
from dataclass_array.typing import DcOrArray, DcT  # pylint: disable=g-multiple-import
from etils import edc
from etils import enp
from etils.array_types import FloatArray  # pylint: disable=g-multiple-import
from visu3d import array_dataclass
from visu3d import plotly
from visu3d.dc_arrays import transformation
from visu3d.utils import np_utils
from visu3d.utils import py_utils
from visu3d.utils.lazy_imports import plotly_base

del abc  # TODO(epot): Why pytype don't like abc ?

assert_supports_protocol = functools.partial(
    py_utils.assert_supports_protocol,
    msg=(
        'Required to support camera transformations. See `v3d.Point3d` and '
        '`v3d.Point2d` for example.'
    ),
)


@edc.dataclass(kw_only=True)
@dataclasses.dataclass(frozen=True)
class TraceConfig(plotly.TraceConfig):
  """Figure configuration.

  Attributes:
    scale: The scale of the camera.
  """

  scale: float = 1.0


@dataclasses.dataclass(frozen=True)
class CameraSpec(array_dataclass.DataclassArray):  # (abc.ABC):
  """Camera intrinsics specification.

  Define the interface of camera model. See `PinholeCamera` for an example of
  class implementation.

  Support batching to allow to stack multiple cameras with the same resolution
  in a single `CameraSpec`.

  ```python
  specs = dca.stack([PinholeCamera(...) for _ in range(10)])

  isinstance(specs, CameraSpec)
  assert specs.shape == (10,)
  assert specs.px_centers().shape == (10, h, w, 2)
  ```

  CameraSpec also allow to project from/to pixel coordinates.

  ```python
  px_coords = specs.px_from_cam @ cam_coords
  ```

  This works with:

  * `xnp.array`: `(..., 3) -> (..., 2)`
  * `v3d.Point3d` -> `v3d.Point2d`
  * Your custom objects. To support this transformation, your dataclass should
    implement the protocols:

    * `apply_px_from_cam(self, camera_spec: v3d.CameraSpec)`:
      to support: `spec.px_from_cam @ my_obj`
    * `apply_cam_from_px(self, camera_spec: v3d.CameraSpec)`:
      to support: `spec.cam_from_px @ my_obj`

  Attributes:
    resolution: Camera resolution (in px).
    h: Camera height resolution (in px).
    w: Camera width resolution (in px).
    fig_config: Additional figure configuration.
  """

  resolution: Tuple[int, int]

  # Overwrite `v3d.DataclassArray.fig_config`.
  fig_config: TraceConfig = dataclasses.field(
      default=TraceConfig(),
      repr=False,
      init=False,
  )

  @property
  def h(self) -> int:
    return self.resolution[0]

  @property
  def w(self) -> int:
    return self.resolution[1]

  @property
  def hw(self) -> tuple[int, int]:
    """`(Height, Width)` in pixel (for usage in `(i, j)` coordinates)."""
    return (self.h, self.w)

  @property
  def wh(self) -> tuple[int, int]:
    """`(Width, Height)` in pixel (for usage in `(u, v)` coordinates)."""
    return (self.w, self.h)

  # @abc.abstractmethod
  @property
  def px_from_cam(self) -> transformation.TransformBase:
    """Project camera 3d coordinates to px 2d coordinates.

    Input can have arbitrary batch shape, including no batch shape for a
    single point as input.

    Returns:
      The transformation 3d cam coordinates -> 2d pixel coordinates.
    """
    raise NotImplementedError

  # @abc.abstractmethod
  @property
  def cam_from_px(self) -> transformation.TransformBase:
    """Unproject 2d pixel coordinates in image space to camera space.

    Note: Points returned by this function are not normalized. Points
    are returned at z=1.

    Input can have arbitrary batch shape, including no batch shape for a
    single point as input.

    Returns:
      The transformation 2d pixel coordinates -> 3d cam coordinates.
    """
    raise NotImplementedError

  # @abc.abstractmethod
  def px_centers(self) -> FloatArray['*shape h w 2']:
    """Returns 2D coordinates of centers of all pixels in the camera image.

    This camera model uses the conventions:

     * Top-left corner of the image is `(0, 0)`
     * Bottom-right corner is `(w, h)` (NOT `(h, w)`)
     * Pixels are centered, so `px_centers()[0, 0] == 0.5`

    Returns:
      2D image coordinates of center of all pixels of shape `(h, w, 2)`.
    """
    raise NotImplementedError

  # Could be removed but only kept for type-checking / auto-complete.
  def replace_fig_config(  # pylint: disable=useless-parent-delegation
      self: DcT,
      *,
      name: str = ...,  # pytype: disable=annotation-type-mismatch
      scale: float = ...,  # pytype: disable=annotation-type-mismatch
      **kwargs: Any,
  ) -> DcT:
    """Returns a copy of self with figure params overwritten."""
    return super().replace_fig_config(name=name, scale=scale, **kwargs)

  # Protocols & internals

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    # TODO(epot): Add arrow to indicates the orientation ?
    start, end = self._get_camera_lines()
    return plotly.make_lines_traces(start=start, end=end)

  @dca.vectorize_method
  def _get_camera_lines(self) -> FloatArray['*shape 4 3']:
    corners_px = [  # Screen corners, in (u, v) coordinates
        [0, 0],
        [0, self.h],
        [self.w, 0],
        [self.w, self.h],
    ]
    corners_world = self.cam_from_px @ self.xnp.array(corners_px)
    corners_world = corners_world * self.fig_config.scale

    start = [
        # 4 lines from center -> corners
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        # 4 lines for the frame
        corners_world[0],
        corners_world[1],
        corners_world[2],
        corners_world[3],
    ]
    end = [
        # 4 lines from center -> corners
        corners_world[0],
        corners_world[1],
        corners_world[2],
        corners_world[3],
        # 4 lines for the frame
        corners_world[1],
        corners_world[3],
        corners_world[0],
        corners_world[2],
    ]
    return self.xnp.asarray(start), self.xnp.asarray(end)


@dataclasses.dataclass(frozen=True)
class PinholeCamera(CameraSpec):
  """Simple camera model.

  Camera conventions:
  In camera/pixel coordinates:

  * u == x == j == w (orientation: →)
  * v == y == i == h (orientation: ↓)

  In camera frame coordinates:

  * `(0, 0, 1)` is at the center of the image
  * x: →
  * y: ↓
  * z is pointing forward

  Attributes:
    K: Camera intrinsics parameters.
    resolution: (h, w) resolution
  """

  K: FloatArray['*shape 3 3']  # pylint: disable=invalid-name

  @classmethod
  def from_focal(
      cls,
      *,
      resolution: tuple[int, int],
      focal_in_px: float,
      xnp: Optional[enp.NpModule] = None,
  ) -> PinholeCamera:
    """Camera factory.

    Args:
      resolution: `(h, w)` resolution in pixel
      focal_in_px: Focal length in pixel
      xnp: `numpy`, `jax.numpy` or `tf.experimental.numpy`. Numpy module to use.
        Default to `numpy`.

    Returns:
      A `PinholeCamera` instance with provided intrinsics.
    """
    if xnp is None:
      xnp = enp.lazy.get_xnp(focal_in_px, strict=False)

    # TODO(epot): Could provide more customizability
    # * Support `focal_in_mm` / `sensor_width`
    # * Support custom central point (cx, cy)
    # * Support different focal for w, h (fx, fy)
    # Warning: Which API when convensions are inconsistents ?
    # resolution == (h, w) but (fx, fy) == (fw, fh)

    # Central point in pixel (offset of the (0, 0) pixel)
    # Because our pixel coordinates are (0, 1), we set the central point
    # to the middle.
    h, w = resolution
    ch = h / 2  # h == y
    cw = w / 2  # w == x

    K = xnp.array(  # pylint: disable=invalid-name
        [
            [focal_in_px, 0, cw],  # cx
            [0, focal_in_px, ch],  # cy
            [0, 0, 1],
        ]
    )
    return cls(
        K=K,
        resolution=resolution,
    )

  @property
  @transformation.custom_transform
  def px_from_cam(  # pylint: disable=bad-staticmethod-argument,property-with-parameters
      self,
      points3d: DcOrArray,
  ) -> DcOrArray:
    points3d = dca.utils.np_utils.asarray(points3d, xnp=self.xnp)
    if isinstance(points3d, dca.DataclassArray):
      assert_supports_protocol(points3d, 'apply_px_from_cam')
      return points3d.apply_px_from_cam(self)
    else:
      return self._px_from_cam(points3d)

  @typing.overload
  def _px_from_cam(
      self,
      points3d: FloatArray['*d 3'],
      *,
      with_depth: bool = False,
  ) -> FloatArray['*d 2']:
    ...

  @typing.overload
  def _px_from_cam(
      self,
      points3d: FloatArray['*d 3'],
      *,
      with_depth: bool = True,
  ) -> tuple[FloatArray['*d 2'], FloatArray['*d 1']]:
    ...

  def _px_from_cam(
      self,
      points3d,
      *,
      with_depth: bool = False,
  ):
    """Project camera 3d coordinates to px 2d coordinates (internal).

    Args:
      points3d: 3d points
      with_depth: If `True`, also return the depth in a separated valiable

    Returns:
      `point2d`: If `with_depth=False`
      `(point2d, depth)`: If `with_depth=True`
    """
    if points3d.shape[-1] != 3:
      raise ValueError(f'Expected cam coords {points3d.shape} to be (..., 3)')

    # K @ [X,Y,Z] -> s * [u, v, 1]
    # (3, 3) @ (..., 3) -> (..., 3)
    points2d = self.xnp.einsum('ij,...j->...i', self.K, points3d)
    # Normalize: s * [u, v, 1] -> [u, v, 1]
    # And only keep [u, v]
    depth = points2d[..., 2:3]
    points2d = points2d[..., :2] / (depth + 1e-8)

    if with_depth:
      return points2d, depth
    else:
      return points2d

  @property
  @transformation.custom_transform
  def cam_from_px(  # pylint: disable=bad-staticmethod-argument,property-with-parameters
      self,
      points2d: FloatArray['*d 2'],
  ) -> FloatArray['*d 3']:
    points2d = dca.utils.np_utils.asarray(points2d, xnp=self.xnp)
    if isinstance(points2d, dca.DataclassArray):
      assert_supports_protocol(points2d, 'apply_cam_from_px')
      return points2d.apply_cam_from_px(self)
    else:
      return self._cam_from_px(points2d)

  def _cam_from_px(  # pylint: disable=bad-staticmethod-argument,property-with-parameters
      self,
      points2d: FloatArray['*d 2'],
  ) -> FloatArray['*d 3']:
    assert not self.shape  # Should be vectorized
    points2d = dca.utils.np_utils.asarray(points2d, xnp=self.xnp)
    if points2d.shape[-1] != 2:
      raise ValueError(f'Expected pixel coords {points2d.shape} to be (..., 2)')

    # [u, v] -> [u, v, 1]
    # Concatenate (..., 2) with (..., 1) -> (..., 3)
    points2d = np_utils.append_row(points2d, 1.0, axis=-1)

    # [X,Y,Z] / s = K-1 @ [u, v, 1]
    # (3, 3) @ (..., 3) -> (..., 3)
    k_inv = enp.linalg.inv(self.K)
    points3d = self.xnp.einsum('ij,...j->...i', k_inv, points2d)

    # TODO(epot): Option to return normalized rays ?
    # Set z to -1
    # [X,Y,Z] -> [X, Y, Z=1]
    points3d = points3d / self.xnp.expand_dims(points3d[..., 2], axis=-1)
    return points3d

  @dca.vectorize_method
  def px_centers(self):
    xnp = self.xnp
    coord_w, coord_h = xnp.meshgrid(
        xnp.arange(self.w),  # w
        xnp.arange(self.h),  # h
        indexing='xy',
    )
    points2d = xnp.stack([coord_w, coord_h], axis=-1)
    points2d = xnp.asarray(points2d, dtype=xnp.float32) + 0.5
    assert points2d.shape == (self.h, self.w, 2)
    return points2d



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for point."""

from __future__ import annotations

import dataclass_array as dca
from etils import enp
import pytest
import visu3d as v3d

# Activate the fixture
set_tnp = enp.testing.set_tnp


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (2,), (2, 3)])
@pytest.mark.parametrize('with_color', [False, True])
def test_point(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    with_color: bool,
):
  init_kwargs = {}
  if with_color:
    init_kwargs['rgb'] = xnp.ones(shape=shape + (3,))

  p = v3d.Point3d(p=xnp.ones(shape=shape + (3,)), **init_kwargs)

  tr = v3d.Transform(R=xnp.eye(3), t=xnp.zeros((3,)))
  p2 = tr @ p
  dca.testing.assert_array_equal(p, p2)

  dca.testing.assert_array_equal(p + [0, 0, 0], p)
  dca.testing.assert_array_equal(p + xnp.array([0, 0, 0]), p)

  p_clipped = p.clip(max=[0.5, 0.5, 5])
  p_clipped_expected = v3d.Point3d(p=xnp.array([0.5, 0.5, 1]), **init_kwargs)
  p_clipped_expected = p_clipped_expected.broadcast_to(shape)
  dca.testing.assert_array_equal(p_clipped, p_clipped_expected)

  # Display should works
  _ = p.fig


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (2,), (2, 3)])
@pytest.mark.parametrize('with_color', [False, True])
@pytest.mark.parametrize('with_depth', [False, True])
def test_point_2d(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    with_color: bool,
    with_depth: bool,
):
  init_kwargs = {}
  if with_color:
    init_kwargs['rgb'] = xnp.ones(shape=shape + (3,))
  if with_depth:
    init_kwargs['depth'] = xnp.ones(shape=shape + (1,))

  p = v3d.Point2d(p=xnp.ones(shape=shape + (2,)), **init_kwargs)

  # Point2d can be projected back and forth to camera
  spec = v3d.PinholeCamera.from_focal(resolution=(4, 4), focal_in_px=35)
  spec = spec.as_xnp(xnp)

  p3d = spec.cam_from_px @ p
  assert isinstance(p3d, v3d.Point3d)

  p2d = spec.px_from_cam @ p3d
  assert isinstance(p2d, v3d.Point2d)

  # Round-trip projection from/to pixel should be a no-op
  if with_depth:
    dca.testing.assert_array_equal(p, p2d)

  # Clip
  p_clipped = p.clip(max=[0.5, 5])
  p_clipped_expected = v3d.Point2d(p=xnp.array([0.5, 1]), **init_kwargs)
  p_clipped_expected = p_clipped_expected.broadcast_to(shape)
  dca.testing.assert_array_equal(p_clipped, p_clipped_expected)

  # Point2d does not support transform (3d)
  tr = v3d.Transform(R=xnp.eye(3), t=xnp.zeros((3,)))
  with pytest.raises(
      NotImplementedError,
      match='does not implement the `.apply_transform',
  ):
    _ = tr @ p

  # Display should works
  _ = p.fig



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ray."""

from __future__ import annotations

import dataclass_array as dca
from etils import enp
from etils.array_types import Array
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
set_tnp = enp.testing.set_tnp

assert not v3d.Ray._dca_tree_map_registered


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (2,), (2, 3)])
def test_ray(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
):
  def _broadcast(x: Array['*d'], shape=shape) -> Array['*d 3']:
    return xnp.broadcast_to(xnp.asarray(x), shape + (3,))

  def _ray_broadcast(
      t: Array['3'],
      d: Array['3'],
      shape=shape,
  ) -> v3d.Ray:
    return v3d.Ray(
        pos=_broadcast(t, shape=shape),
        dir=_broadcast(d, shape=shape),
    )

  def _assert_ray_match(p, t, d, shape=shape):
    assert p.xnp is xnp
    assert p.shape == shape
    dca.testing.assert_allclose(p, _ray_broadcast(t=t, d=d, shape=shape))

  t = xnp.array([1, 0, 0])
  d = xnp.array([0, 2, 2])
  sqrt8 = np.sqrt(0**2 + 2**2 + 2**2.0)

  p = _ray_broadcast(t=t, d=d)

  p_brodcast = v3d.Ray(pos=t, dir=d).map_field(_broadcast)
  dca.testing.assert_allclose(p, p_brodcast)

  _assert_ray_match(p.mean(), t=t, d=d, shape=())
  _assert_ray_match(p + [3, 2, -1], t=[4, 2, -1], d=d)
  _assert_ray_match(p.scale_dir(3), t=t, d=[0, 6, 6])
  _assert_ray_match(p.look_at([-2, 3, 1]), t=t, d=[-3, 3, 1])
  _assert_ray_match(p.normalize(), t=t, d=[0, 2 / sqrt8, 2 / sqrt8])

  norm = p.norm()
  assert norm.shape == shape
  np.testing.assert_allclose(
      norm,
      xnp.broadcast_to(sqrt8, shape),
  )

  end = p.end
  assert end.shape == shape + (3,)
  np.testing.assert_allclose(end, _broadcast(end))

  _ = p.fig

  p = p.replace_fig_config(name='new_name')
  assert p.fig_config.name == 'new_name'


@enp.testing.parametrize_xnp()
def test_ray_mean(xnp: enp.NpModule):
  ray = v3d.Ray(
      pos=[0, 0, 0],
      dir=[1, 1, 1],
  )
  ray = ray.as_xnp(xnp)
  ray = ray.broadcast_to((1, 2, 3, 4))

  assert ray.mean().shape == ()  # pylint: disable=g-explicit-bool-comparison
  assert ray.mean(axis=None).shape == ()  # pylint: disable=g-explicit-bool-comparison
  assert ray.mean(axis=0).shape == (2, 3, 4)
  assert ray.mean(axis=1).shape == (1, 3, 4)
  assert ray.mean(axis=2).shape == (1, 2, 4)
  assert ray.mean(axis=3).shape == (1, 2, 3)
  assert ray.mean(axis=-1).shape == (1, 2, 3)
  assert ray.mean(axis=-4).shape == (2, 3, 4)
  assert ray.mean(axis=(0, -1)).shape == (2, 3)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python utils."""

from __future__ import annotations

import dataclasses
import importlib
import types
from typing import Any, Optional, Type, TypeVar, Union

# If these are reused, we could move them to `epy`.

_T = TypeVar('_T')
_SelfOrCls = Union[Any, Type[Any]]


def supports_protocol(
    self_or_cls: _SelfOrCls,
    protocol_fn_name: str,
) -> bool:
  """Returns `True` if the class support the protocol."""
  return hasattr(self_or_cls, protocol_fn_name) and callable(
      getattr(self_or_cls, protocol_fn_name)
  )


# Is it possible to dynamically add TypeGuard+Protocol annotations ? (PEP 647)
def assert_supports_protocol(
    self_or_cls: _SelfOrCls,
    protocol_fn_name: str,
    *,
    msg: str = '',
) -> None:
  """Returns `True` if the class support the protocol."""
  if not supports_protocol(self_or_cls, protocol_fn_name):
    cls = _get_class(self_or_cls)
    raise NotImplementedError(
        f'`{cls.__qualname__}` does not implement the `.{protocol_fn_name}()` '
        f'protocol. {msg}'
    )


def _get_class(self_or_cls: Union[_T, Type[_T]]) -> Type[_T]:
  if isinstance(self_or_cls, type):
    return self_or_cls
  else:
    return type(self_or_cls)


@dataclasses.dataclass
class LazyModule:
  """Module loaded lazily during first call."""

  module_name: str
  module: Optional[types.ModuleType] = None

  def __getattr__(self, name: str) -> Any:
    if self.module is None:  # Load on first call
      self.module = importlib.import_module(self.module_name)
    return getattr(self.module, name)
